from .synthetic_generator import synthetic_generator_config
from .binary_classifier import binary_classifer_config
from .rag_scoring import rag_scoring_config
from .ues_idp import ues_idp_config
from .kilt_filter import KILT_dataset_process
from .superglue_filter import superGlue
from .prompts import context_relevance_system_prompt, answer_relevance_system_prompt, answer_faithfulness_system_prompt
from typing import List
import pandas as pd

class ARES:
    """
    ARES class encapsulates configurations for various components of the ARES system.
    It allows for the setup and management of configurations for synthetic query generation,
    binary classification, RAG scoring, and UES/IDP processes.
    """
    
    # A placeholder object used to denote default values in the configuration specification
    REQUIRED_BUT_HAS_DEFAULT = object()

    # Configuration specification dictionary defining the parameters for each component
    config_spec = {
        "ues_idp": {
            "in_domain_prompts_dataset": (str, None),  # Required parameter with no default value
            "unlabeled_evaluation_set": (str, None),  # Required parameter with no default value
            "context_relevance_system_prompt": (str,context_relevance_system_prompt),  # Optional with default
            "answer_relevance_system_prompt": (str, answer_relevance_system_prompt),  # Optional with default
            "answer_faithfulness_system_prompt": (str, answer_faithfulness_system_prompt),  # Optional with default
            "debug_mode": (bool, False),  # Optional with default
            "documents": (int, 0),  # Optional with default
            "model_choice": (str, "gpt-3.5-turbo-1106"),  # Optional with default
            "request_delay": (int, 0),  # Optional with default
            "vllm": (bool, False),  # Optional with default
            "host_url": (str, "None")  # Optional with default
        },

        "synthetic_query_generator": {
            "document_filepaths": (list, None),  # Required parameter with no default value
            "few_shot_prompt_filename": (str, None),  # Required parameter with no default value
            "synthetic_queries_filenames": (list, None),  # Required parameter with no default value
            "documents_sampled": (int, None),  # Required parameter with no default value
            "model_choice": (str, "google/flan-t5-xxl"),  # Optional with default
            "clean_documents": (bool, False),  # Optional with default
            "regenerate_synth_questions": (bool, True),  # Optional with default
            "percentiles": (list, [0.05, 0.25, 0.5, 0.95]),  # Optional with default
            "question_temperatures": (list, [2.0, 1.5, 1.0, 0.5, 0.0]),  # Optional with default
            "regenerate_answers": (bool, True),  # Optional with default
            "number_of_negatives_added_ratio": (float, 0.5),  # Optional with default
            "lower_bound_for_negatives": (int, 5),  # Optional with default
            "number_of_contradictory_answers_added_ratio": (float, 0.67),  # Optional with default
            "number_of_positives_added_ratio": (float, 0.0),  # Optional with default
            "regenerate_embeddings": (float, True),  # Optional with default
            "synthetic_query_prompt": (str, "You are an expert question-answering system. You must create a question for the provided document. The question must be answerable within the context of the document.\n\n")  # Optional with default
        },
        
        "classifier_model": {
            "training_dataset": (list, None),  # Required parameter with no default value
            "validation_set": (list, None),  # Required parameter with no default value
            "label_column": (list, None),  # Required parameter with no default value
            "num_epochs": (int, None),  # Required parameter with no default value
            "patience_value": (int, None),  # Required parameter with no default value
            "learning_rate": (float, None),  # Required parameter with no default value
            "training_dataset_path": (str, "None"),  # Optional with default
            "validation_dataset_path": (str, "None"),  # Optional with default
            "model_choice": (str, "microsoft/deberta-v3-large"),  # Optional with default
            "validation_set_scoring": (bool, True),  # Optional with default
            "assigned_batch_size": (int, REQUIRED_BUT_HAS_DEFAULT, 1),  # Default is 1
            "gradient_accumulation_multiplier": (int, REQUIRED_BUT_HAS_DEFAULT, 32),  # Default is 32
            "number_of_runs": (int, 1),  # Optional with default
            "num_warmup_steps": (int, 100),  # Optional with default
            "training_row_limit": (int, -1),  # Optional with default
            "validation_row_limit": (int, -1)  # Optional with default
        },

        "ppi": {
            "evaluation_datasets": (list, None),  # Required parameter with no default value
            "labels": (list, None),  # Required parameter with no default value
            "checkpoints": (list, []),  # Required parameter with no default value
            "few_shot_examples_filepath": (str, "None"), # Optional with default
            "gold_label_path": (str, "None"),  # Optional with default
            "rag_type": (str, "question_answering"),  # Optional with default
            "model_choice": (str, "microsoft/deberta-v3-large"),  # Optional with default
            "llm_judge": (str, "None"),  # Optional with default
            "assigned_batch_size": (int, 1),  # Optional with default
            "number_of_labels": (int, 2),  # Optional with default
            "alpha": (int, 0.05),  # Optional with default
            "num_trials": (int, 1000),  # Optional with default
            "vllm": (bool, False),  # Optional with default
            "host_url": (str, "http://0.0.0.0:8000/v1"),  # Optional with default
            "request_delay": (int, 0),  # Optional with default
            "debug_mode": (bool, False),  # Optional with default
            "machine_label_llm_model": (str, "None"),  # Optional with default
            "gold_machine_label_path": (str, "None"),  # Optional with default
            "prediction_filepath": (str, "None")  # Optional with default
        }
    }

    def __init__(self, synthetic_query_generator={}, ues_idp={}, classifier_model={}, ppi={}):
        """
        Initializes the ARES class with configurations for different components.

        Args:
            synthetic_query_generator (dict): Configuration for the synthetic query generator.
            ues_idp (dict): Configuration for UES IDP.
            classifier_model (dict): Configuration for the classifier model.
            ppi (dict): Configuration for PPI (Protein-Protein Interaction).

        Each configuration dictionary is passed to the `prepare_config` method to validate
        and prepare the final configuration based on the default settings defined in `config_spec`.
        """
        self.synthetic_query_generator_config = self.prepare_config("synthetic_query_generator", synthetic_query_generator)
        self.classifier_model_config = self.prepare_config("classifier_model", classifier_model)
        self.ppi_config = self.prepare_config("ppi", ppi)
        self.ues_idp_config = self.prepare_config("ues_idp", ues_idp)

    def generate_synthetic_data(self):
        """
        Generates synthetic data using the synthetic query generator configuration.
        If the configuration is empty, the generation process is skipped.
        """
        if not self.synthetic_query_generator_config:
            print("Skipping synthetic generator configuration due to missing parameters.")
        else:
            synthetic_generator_config(**self.synthetic_query_generator_config)

    def train_classifier(self):
        """
        Trains the classifier using the classifier model configuration.
        If the configuration is empty, the training process is skipped.
        """
        if not self.classifier_model_config:
            print("Skipping binary classifier configuration due to missing parameters.")
        else:
            binary_classifer_config(**self.classifier_model_config)

    def evaluate_RAG(self):
        """
        Evaluates the RAG using the PPI configuration.
        If the configuration is empty, the evaluation process is skipped.
        """
        if not self.ppi_config:
            print("Skipping RAG evaluation configuration due to no parameters")
        else:
            return rag_scoring_config(**self.ppi_config)

    def ues_idp(self):
        """
        Executes the UES IDP configuration.
        """
        return ues_idp_config(**self.ues_idp_config)

    def KILT_dataset(self, dataset_name):
        """
        Processes the KILT dataset with the specified dataset name.
        """
        KILT_dataset_process(dataset_name)

    def superGlue_dataset(self, dataset_name):
        """
        Processes the SuperGlue dataset with the specified dataset name.
        """
        superGlue(dataset_name)

    def run(self):
        """
        Executes the configurations for synthetic data generation, classifier training, and RAG evaluation.
        If any configuration is empty, the corresponding process is skipped.
        """
        self.generate_synthetic_data()
        self.train_classifier()
        self.evaluate_RAG()

        if not self.ues_idp_config:
            print(f"Skipping UES and IDP since either unlabeled evaluation set or in domain prompts not provided")
            exit()
        else:
            model_choice = self.ues_idp_config["model_choice"]
            ues_idp_config(**self.ues_idp_config)

    def prepare_config(self, component_name, user_config):
        """
        Prepares and validates the configuration for a given component based on the default settings.

        Args:
            component_name (str): The name of the component for which the configuration is being prepared.
            user_config (dict): The user-provided configuration dictionary.

        Returns:
            dict: The prepared and validated configuration dictionary.

        Raises:
            TypeError: If the provided value type does not match the expected type.
            ValueError: If a required parameter is missing and no default is specified.
        """
        if not user_config:
            return {}
        component = self.config_spec[component_name]
        prepared_config = {}
        for param, config in component.items():
            expected_type, default, default_value = config if len(config) == 3 else (*config, None)
            if param in user_config:
                value = user_config[param]
                if not isinstance(value, expected_type):
                    raise TypeError(f"Parameter '{param}' for {component_name} is expected to be of type {expected_type.__name__}, received {type(value).__name__} instead.")
                prepared_config[param] = value
            elif default is self.REQUIRED_BUT_HAS_DEFAULT:
                # If the parameter is required but not provided, use the specified default value.
                if param not in user_config:
                    print(f"\nWarning: '{param}' not provided for {component_name}, using default value {default_value}.")
                prepared_config[param] = default_value
            elif default is not None:
                prepared_config[param] = default
            else:
                raise ValueError(f"Missing required parameter '{param}' for {component_name}.")
        return prepared_config


import numpy as np
import matplotlib.pyplot as plt
import itertools
from scipy.stats import binom, norm
from scipy.special import expit
from scipy.optimize import brentq
from sklearn.linear_model import LogisticRegression
from joblib import delayed, Parallel
import pdb

"""
    IID Concentration Bounds
"""
def binomial_iid(N: int, alpha: float, muhat: float) -> np.ndarray:
    """
    Calculate the binomial confidence interval for independent and identically distributed (IID) samples.

    Parameters:
    N (int): The number of trials.
    alpha (float): The significance level.
    muhat (float): The estimated mean.

    Returns:
    np.ndarray: A numpy array containing the lower and upper bounds of the confidence interval.
    """
    
    def invert_upper_tail(mu: float) -> float:
        """Calculate the upper tail inversion."""
        return binom.cdf(N * muhat, N, mu) - (alpha / 2)
    
    def invert_lower_tail(mu: float) -> float:
        """Calculate the lower tail inversion."""
        return binom.cdf(N * muhat, N, mu) - (1 - alpha / 2)
    
    u = brentq(invert_upper_tail, 0, 1)
    l = brentq(invert_lower_tail, 0, 1)
    
    return np.array([l, u])

def bentkus_iid(N: int, alpha: float, muhat: float) -> np.ndarray:
    """
    Calculate the Bentkus confidence interval for independent and identically distributed (IID) samples.

    Parameters:
    N (int): The number of trials.
    alpha (float): The significance level.
    muhat (float): The estimated mean.

    Returns:
    np.ndarray: A numpy array containing the lower and upper bounds of the confidence interval.
    """
    return binomial_iid(N, alpha / np.e, muhat)

def clt_iid(x: np.ndarray, alpha: float) -> np.ndarray:
    """
    Calculate the confidence interval for the mean of IID samples using the Central Limit Theorem (CLT).

    Parameters:
    x (np.ndarray): The array of sample data.
    alpha (float): The significance level.

    Returns:
    np.ndarray: A numpy array containing the lower and upper bounds of the confidence interval.
    """
    n = x.shape[0]  # Number of samples
    sigmahat = x.std()  # Sample standard deviation
    w = norm.ppf(1 - alpha / 2) * sigmahat / np.sqrt(n)  # Margin of error
    return np.array([x.mean() - w, x.mean() + w])  # Confidence interval

def wsr_iid(x_n: np.ndarray, alpha: float, grid: np.ndarray, num_cpus: int = 10, parallelize: bool = False, 
            intersection: bool = True, theta: float = 0.5, c: float = 0.75) -> np.ndarray:
    """
    Calculate the Weighted Sequential Rank (WSR) confidence interval for IID samples.

    Parameters:
    x_n (np.ndarray): The array of sample data.
    alpha (float): The significance level.
    grid (np.ndarray): The grid of values to search for the confidence interval.
    num_cpus (int, optional): The number of CPUs to use for parallel processing. Default is 10.
    parallelize (bool, optional): Whether to parallelize the computation. Default is False.
    intersection (bool, optional): Whether to use the intersection method for the confidence interval. Default is True.
    theta (float, optional): The weighting parameter. Default is 0.5.
    c (float, optional): The scaling parameter. Default is 0.75.

    Returns:
    np.ndarray: A numpy array containing the lower and upper bounds of the confidence interval.
    """
    n = x_n.shape[0]
    t_n = np.arange(1, n + 1)
    muhat_n = (0.5 + np.cumsum(x_n)) / (1 + t_n)
    sigma2hat_n = (0.25 + np.cumsum(np.power(x_n - muhat_n, 2))) / (1 + t_n)
    sigma2hat_tminus1_n = np.append(0.25, sigma2hat_n[:-1])
    assert np.all(sigma2hat_tminus1_n > 0), "All elements of sigma2hat_tminus1_n must be greater than 0"
    lambda_n = np.sqrt(2 * np.log(2 / alpha) / (n * sigma2hat_tminus1_n))

    def M(m: float) -> np.ndarray:
        lambdaplus_n = np.minimum(lambda_n, c / m)
        lambdaminus_n = np.minimum(lambda_n, c / (1 - m))
        return np.maximum(
            theta * np.exp(np.cumsum(np.log(1 + lambdaplus_n * (x_n - m)))),
            (1 - theta) * np.exp(np.cumsum(np.log(1 - lambdaminus_n * (x_n - m))))
        )

    if parallelize:
        M_vectorized = np.vectorize(M)
        M_list = Parallel(n_jobs=num_cpus)(delayed(M_vectorized)(m) for m in grid)
        indicators_gxn = np.array(M_list) < 1 / alpha
    else:
        indicators_gxn = np.zeros([grid.size, n])
        found_lb = False
        for m_idx, m in enumerate(grid):
            m_n = M(m)
            indicators_gxn[m_idx] = m_n < 1 / alpha
            if not found_lb and np.prod(indicators_gxn[m_idx]):
                found_lb = True
            if found_lb and not np.prod(indicators_gxn[m_idx]):
                break  # Since interval, once find a value that fails, stop searching

    if intersection:
        ci_full = grid[np.where(np.prod(indicators_gxn, axis=1))[0]]
    else:
        ci_full = grid[np.where(indicators_gxn[:, -1])[0]]

    if ci_full.size == 0:  # Grid may be too coarse
        idx = np.argmax(np.sum(indicators_gxn, axis=1))
        if idx == 0:
            return np.array([grid[0], grid[1]])
        return np.array([grid[idx - 1], grid[idx]])

    return np.array([ci_full.min(), ci_full.max()])  # Only output the interval

"""
    Mean estimation confidence intervals
"""

def pp_mean_iid_asymptotic(Y_labeled: np.ndarray, Yhat_labeled: np.ndarray, Yhat_unlabeled: np.ndarray, alpha: float) -> list:
    """
    Compute the mean estimation confidence interval for IID data using an asymptotic approach.

    Parameters:
    Y_labeled (np.ndarray): Labeled data array.
    Yhat_labeled (np.ndarray): Predicted values for the labeled data.
    Yhat_unlabeled (np.ndarray): Predicted values for the unlabeled data.
    alpha (float): Significance level for the confidence interval.

    Returns:
    list: A list containing the lower and upper bounds of the confidence interval.
    """
    n = Y_labeled.shape[0]  # Number of labeled samples
    N = Yhat_unlabeled.shape[0]  # Number of unlabeled samples
    
    if n == 0 or N == 0:
        return [0, 0]

    # Mean of the predicted values for the unlabeled data
    tildethetaf = Yhat_unlabeled.mean()

    # Mean of the residuals (difference between predicted and actual values for labeled data)
    rechat = (Yhat_labeled - Y_labeled).mean()
    
    # Point estimate for the parameter of interest
    thetahatPP = tildethetaf - rechat

    # Standard deviation of the predicted values for the unlabeled data
    sigmaftilde = np.std(Yhat_unlabeled)

    # Standard deviation of the residuals
    sigmarec = np.std(Yhat_labeled - Y_labeled)

    # Half-width of the confidence interval
    hw = norm.ppf(1 - alpha / 2) * np.sqrt((sigmaftilde**2 / N) + (sigmarec**2 / n))

    # Return the confidence interval as a list
    return [thetahatPP - hw, thetahatPP + hw]

"""
    OLS algorithm with sandwich variance estimator
"""

def ols(features: np.ndarray, outcome: np.ndarray) -> np.ndarray:
    """
    Perform Ordinary Least Squares (OLS) regression.

    Parameters:
    features (np.ndarray): The input feature matrix (design matrix).
    outcome (np.ndarray): The outcome vector (response variable).

    Returns:
    np.ndarray: The OLS coefficients.
    """
    ols_coeffs = np.linalg.pinv(features).dot(outcome)
    return ols_coeffs

def classical_ols_interval(X: np.ndarray, Y: np.ndarray, alpha: float, 
                           return_stderr: bool = False, sandwich: bool = True):
    """
    Compute the confidence interval for OLS regression coefficients using the sandwich variance estimator.

    Parameters:
    X (np.ndarray): The input feature matrix (design matrix).
    Y (np.ndarray): The outcome vector (response variable).
    alpha (float): Significance level for the confidence interval.
    return_stderr (bool): If True, return the standard errors instead of the confidence interval. Default is False.
    sandwich (bool): If True, use the sandwich variance estimator. Default is True.

    Returns:
    Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]: The standard errors if return_stderr is True, otherwise the lower and upper bounds of the confidence interval.
    """
    n = X.shape[0]  # Number of samples
    thetahat = ols(X, Y)  # OLS coefficients
    Sigmainv = np.linalg.inv(1/n * X.T @ X)  # Inverse of the covariance matrix

    if sandwich:
        M = 1/n * (X.T * ((Y - X @ thetahat) ** 2)[None, :]) @ X
    else:
        M = 1/n * ((Y - X @ thetahat) ** 2).mean() * X.T @ X

    V = Sigmainv @ M @ Sigmainv  # Variance-covariance matrix
    stderr = np.sqrt(np.diag(V))  # Standard errors

    if return_stderr:
        return stderr

    halfwidth = norm.ppf(1 - alpha / 2) * stderr / np.sqrt(n)  # Half-width of the confidence interval
    return thetahat - halfwidth, thetahat + halfwidth  # Confidence interval

def pp_ols_interval(X_labeled: np.ndarray, X_unlabeled: np.ndarray, Y_labeled: np.ndarray, 
                    Yhat_labeled: np.ndarray, Yhat_unlabeled: np.ndarray, alpha: float, 
                    sandwich: bool = True) -> tuple[np.ndarray, np.ndarray]:
    """
    Compute the confidence interval for post-processed OLS regression coefficients.

    Parameters:
    X_labeled (np.ndarray): The labeled input feature matrix (design matrix).
    X_unlabeled (np.ndarray): The unlabeled input feature matrix (design matrix).
    Y_labeled (np.ndarray): The labeled outcome vector (response variable).
    Yhat_labeled (np.ndarray): The predicted outcomes for the labeled data.
    Yhat_unlabeled (np.ndarray): The predicted outcomes for the unlabeled data.
    alpha (float): Significance level for the confidence interval.
    sandwich (bool): If True, use the sandwich variance estimator. Default is True.

    Returns:
    Tuple[np.ndarray, np.ndarray]: The lower and upper bounds of the confidence interval.
    """
    n = X_labeled.shape[0]  # Number of labeled samples
    N = X_unlabeled.shape[0]  # Number of unlabeled samples

    # Compute OLS coefficients for the unlabeled data
    thetatildef = ols(X_unlabeled, Yhat_unlabeled)
    
    # Compute the rectifier OLS coefficients for the labeled data
    rectifierhat = ols(X_labeled, Y_labeled - Yhat_labeled)
    
    # Post-processed OLS coefficients
    pp_thetahat = thetatildef + rectifierhat
    
    # Compute standard errors for the unlabeled and labeled data
    stderr_tildef = classical_ols_interval(X_unlabeled, Yhat_unlabeled, 0.001 * alpha, return_stderr=True, sandwich=sandwich)
    stderr_rec = classical_ols_interval(X_labeled, Y_labeled - Yhat_labeled, 0.999 * alpha, return_stderr=True, sandwich=sandwich)
    
    # Compute the half-width of the confidence interval
    halfwidth = norm.ppf(1 - alpha / 2) * np.sqrt((stderr_rec**2 / n) + (stderr_tildef**2 / N))
    
    # Return the confidence interval
    return pp_thetahat - halfwidth, pp_thetahat + halfwidth

"""
    Logistic regression algorithm
"""
def logistic(X: np.ndarray, y: np.ndarray) -> np.ndarray:
    """
    Perform logistic regression on the given data.

    Parameters:
    X (np.ndarray): The input feature matrix (design matrix).
    y (np.ndarray): The binary outcome vector (response variable).

    Returns:
    np.ndarray: The coefficients of the logistic regression model.
    """
    # Initialize and fit the logistic regression model
    clf = LogisticRegression(
        penalty='none', 
        solver='lbfgs', 
        max_iter=10000, 
        tol=1e-15, 
        fit_intercept=False
    ).fit(X, y)
    
    # Return the coefficients of the fitted model
    return clf.coef_.squeeze()

def product(*args: tuple, **kwds: dict) -> iter:
    """
    Cartesian product of input iterables.

    Parameters:
    *args (tuple): Variable length argument list of input iterables.
    **kwds (dict): Variable length keyword arguments (not used in this function).

    Yields:
    iter: An iterator over tuples representing the Cartesian product of the input iterables.

    Examples:
    ---------
    product('ABCD', 'xy') --> Ax Ay Bx By Cx Cy Dx Dy
    product(range(2), repeat=3) --> 000 001 010 011 100 101 110 111
    """
    pools = map(tuple, args)  # Convert each input iterable to a tuple
    result = [[]]  # Initialize the result with an empty list

    # Generate the Cartesian product
    for pool in pools:
        result = [x + [y] for x in result for y in pool]

    # Yield each product as a tuple
    for prod in result:
        yield tuple(prod)

def classical_logistic_interval(X: np.ndarray, Y: np.ndarray, alpha: float, num_grid: int = 500) -> list:
    """
    Calculate the classical logistic confidence interval.

    Parameters:
    X (np.ndarray): The input feature matrix (design matrix).
    Y (np.ndarray): The binary outcome vector (response variable).
    alpha (float): The significance level for the confidence interval.
    num_grid (int, optional): The number of grid points to use for the interval calculation. Default is 500.

    Returns:
    list: A list containing the minimum and maximum values of the confidence interval.
    """
    n = X.shape[0]  # Number of samples
    d = X.shape[1]  # Number of features
    Y = (Y >= 0.5).astype(int)  # Convert Y to binary (0 or 1)
    
    # Obtain the point estimate using logistic regression
    point_estimate = logistic(X, Y)

    # Create a grid of theta values around the point estimate
    theta_grid = np.concatenate([
        np.linspace(-3 * point_estimate, point_estimate, num_grid // 2),
        np.linspace(point_estimate, 3 * point_estimate, num_grid // 2)[1:]
    ])

    # Calculate the expected value of the logistic function
    mu = expit(X @ theta_grid.T)
    
    # Calculate the gradient
    g = 1 / n * X.T @ (mu - Y[:, None])

    # Estimate the standard error
    sigmahat_err = np.std(X[:, :, None] * (mu - Y[:, None])[:, None, :], axis=0)
    
    # Calculate the half-width of the gradient confidence interval
    grad_halfwidth = norm.ppf(1 - alpha / (2 * d)) * sigmahat_err / np.sqrt(n)

    # Determine the condition for the confidence interval
    condition = np.all(np.abs(g) <= grad_halfwidth, axis=0)

    # Extract the confidence interval points
    Cpp = theta_grid[condition]

    # TODO: If all positive, make grid wider
    
    # Ensure the condition is met at the boundaries
    assert (condition[0] == False) & (condition[-1] == False)

    # Return the minimum and maximum values of the confidence interval
    return [Cpp.min(axis=0), Cpp.max(axis=0)]

def pp_logistic_interval(X_labeled: np.ndarray, X_unlabeled: np.ndarray, Y_labeled: np.ndarray, 
                         Yhat_labeled: np.ndarray, Yhat_unlabeled: np.ndarray, alpha: float, 
                         num_grid: int = 500) -> list:
    """
    Calculate the logistic confidence interval using both labeled and unlabeled data.

    Parameters:
    X_labeled (np.ndarray): The labeled input feature matrix.
    X_unlabeled (np.ndarray): The unlabeled input feature matrix.
    Y_labeled (np.ndarray): The binary outcome vector for labeled data.
    Yhat_labeled (np.ndarray): The predicted probabilities for labeled data.
    Yhat_unlabeled (np.ndarray): The predicted probabilities for unlabeled data.
    alpha (float): The significance level for the confidence interval.
    num_grid (int, optional): The number of grid points to use for the interval calculation. Default is 500.

    Returns:
    list: A list containing the minimum and maximum values of the confidence interval.
    """
    
    # Combine labeled and unlabeled feature matrices
    X = np.concatenate([X_labeled, X_unlabeled], axis=0)
    
    # Number of labeled samples
    n = X_labeled.shape[0]
    
    # Number of features
    d = X_labeled.shape[1]
    
    # Number of unlabeled samples
    N = X_unlabeled.shape[0]
    
    # Clip predicted probabilities to be within [0, 1]
    Yhat_labeled = np.clip(Yhat_labeled, 0, 1)
    Yhat_unlabeled = np.clip(Yhat_unlabeled, 0, 1)

    # Combine predicted probabilities
    Yhat = np.concatenate([Yhat_labeled, Yhat_unlabeled], axis=0)
    
    # Obtain the point estimate using logistic regression on labeled data
    point_estimate = logistic(X_labeled, (Y_labeled > 0.5).astype(int))

    # Calculate the rechat term
    rechat = 1/n * X_labeled.T @ (Yhat_labeled - Y_labeled)
    
    # Estimate the standard error for the rechat term
    sigmahat_rec = np.std(X_labeled * (Yhat_labeled - Y_labeled)[:, None], axis=0)

    # Create a grid of theta values around the point estimate
    theta_grid = np.concatenate([
        np.linspace(-3 * point_estimate, point_estimate, num_grid // 2),
        np.linspace(point_estimate, 3 * point_estimate, num_grid // 2)[1:]
    ])

    # Calculate the expected value of the logistic function for the unlabeled data
    mu = expit(X_unlabeled @ theta_grid.T)
    
    # Calculate the gradient for the unlabeled data
    g = 1/N * X_unlabeled.T @ (mu - Yhat_unlabeled[:, None])

    # Estimate the standard error for the gradient
    sigmahat_err = np.std(X_unlabeled[:, :, None] * (mu - Yhat_unlabeled[:, None])[:, None, :], axis=0)

    # Calculate the half-width of the confidence interval
    halfwidth = norm.ppf(1 - alpha / (2 * d)) * np.sqrt(sigmahat_rec[:, None]**2 / n + sigmahat_err**2 / N)

    # Determine the condition for the confidence interval
    condition = np.all(np.abs(g + rechat[:, None]) <= halfwidth, axis=0)

    # Extract the confidence interval points
    Cpp = theta_grid[condition]

    # TODO: If all positive, make grid wider
    assert (condition[0] == False) & (condition[-1] == False)

    # Return the minimum and maximum values of the confidence interval
    return [Cpp.min(axis=0), Cpp.max(axis=0)]

"""
    DISCRETE L_p ESTIMATION RATES
"""

def linfty_dkw(N: int, K: int, alpha: float) -> float:
    """
    Calculate the L-infinity Dvoretzky-Kiefer-Wolfowitz (DKW) bound.

    Parameters:
    N (int): The number of samples.
    K (int): The number of bins or categories.
    alpha (float): The significance level.

    Returns:
    float: The L-infinity DKW bound.
    """
    return np.sqrt(2 / N * np.log(2 / alpha))

def linfty_binom(N: int, K: int, alpha: float, qhat: np.ndarray) -> float:
    """
    Calculate the L-infinity binomial bound.

    Parameters:
    N (int): The number of samples.
    K (int): The number of bins or categories.
    alpha (float): The significance level.
    qhat (np.ndarray): The estimated probabilities for each bin.

    Returns:
    float: The L-infinity binomial bound.
    """
    epsilon = 0
    for k in np.arange(K):
        bci = binomial_iid(N, alpha / K, qhat[k])
        epsilon = np.maximum(epsilon, np.abs(bci - qhat[k]).max())
    return epsilon
"""
	SAMPLING WITHOUT REPLACEMENT
"""
def clt_swr(x: np.ndarray, N: int, alpha: float) -> np.ndarray:
    """
    Calculate the confidence interval for the mean of a sample without replacement using the Central Limit Theorem (CLT).

    Parameters:
    x (np.ndarray): The sample data as a numpy array.
    N (int): The total population size.
    alpha (float): The significance level.

    Returns:
    np.ndarray: A numpy array containing the lower and upper bounds of the confidence interval.
    """
    n = x.shape[0]  # Sample size
    point_estimate = x.mean()  # Mean of the sample
    # Standard deviation of the sample adjusted for finite population correction
    fluctuations = x.std() * norm.cdf(1 - alpha / 2) * np.sqrt((N - n) / (N * n))
    # Return the confidence interval as a numpy array
    return np.array([point_estimate - fluctuations, point_estimate + fluctuations])

def wsr_swr(x: np.ndarray, N: int, alpha: float, grid: np.ndarray, num_cpus: int = 10, intersection: bool = True) -> np.ndarray:
    """
    Calculate the confidence interval for a bounded sequence using weighted sampling without replacement.

    Parameters:
    x (np.ndarray): A [0,1] bounded sequence.
    N (int): The total population size.
    alpha (float): The significance level.
    grid (np.ndarray): The grid of values to evaluate.
    num_cpus (int, optional): The number of CPUs to use for parallel processing. Default is 10.
    intersection (bool, optional): Whether to use intersection in the calculation. Default is True.

    Returns:
    np.ndarray: A numpy array containing the lower and upper bounds of the confidence interval.
    """
    n = x.shape[0]  # Sample size

    def mu(m: float, i: int) -> np.ndarray:
        """Calculate the mean adjusted for finite population correction."""
        return (N * m - np.concatenate([np.array([0]), np.cumsum(x[:i-1])])) / (N - (np.arange(i) + 1) + 1)

    muhats = (1/2 + np.cumsum(x)) / (np.arange(n) + 1)  # Estimated means
    sigmahat2s = (1/4 + np.cumsum((x - muhats) ** 2)) / (np.arange(n) + 1)  # Estimated variances
    lambdas = np.concatenate([np.array([1]), np.sqrt(2 * np.log(2 / alpha) / (n * sigmahat2s))[:-1]])  # Lambda values

    def M(m: float, i: int) -> float:
        """Calculate the M value for the given m and i."""
        return 1/2 * np.maximum(
            np.prod(1 + np.minimum(lambdas[:i], 0.5 / mu(m, i)) * (x[:i] - mu(m, i))),
            np.prod(1 - np.minimum(lambdas[:i], 0.5 / (1 - mu(m, i))) * (x[:i] - mu(m, i)))
        )

    M = np.vectorize(M)  # Vectorize the M function for parallel processing

    if intersection:
        M_list = Parallel(n_jobs=num_cpus)(delayed(M)(grid, i) for i in range(1, n + 1))
    else:
        M_list = [M(grid, n)]

    ci_full = grid[np.where(np.prod(np.stack(M_list, axis=1) < 1 / alpha, axis=1))[0]]
    return np.array([ci_full.min(), ci_full.max()])  # Only output the interval
